import pandas as pd
import numpy as np
import os

from d3m.container.pandas import DataFrame
from d3m.container.dataset import Dataset
from load_d3m_dataset import load_dataset
from d3m.primitives.data_preprocessing.random_sampling_imputer import BYU as RandomSamplingImputer

def print_missing_vals_info(df, df_name):
    num_empty_cols = 0
    total_num_nan = 0
    for feature in df:
        num_nan = np.sum(df[feature].isnull())
        if num_nan == len(df[feature]):
            num_empty_cols += 1
        else:
            total_num_nan += num_nan
    print('\nMissing Values for ', df_name)
    print('\nEmpty columns: ', num_empty_cols)
    print('Total missing values (not counting empty columns): ', total_num_nan)

if __name__ == '__main__':
    df = load_dataset()
    df.replace("", np.nan, inplace=True)
    print_missing_vals_info(df, 'Input Dataset')
    imputer = RandomSamplingImputer(hyperparams=None, random_seed=0)
    imputer.set_training_data(inputs=df)
    imputer.fit()
    new_df = imputer.produce(inputs=df).value
    print_missing_vals_info(new_df, 'Imputed Dataset')


import pandas as pd
import os
import json
import numpy as np
import pprint

from d3m.container.dataset import Dataset
from d3m.container.pandas import DataFrame
from d3m.metadata import base as metadata_base
from load_d3m_dataset import load_dataset

from d3m.primitives.metafeature_extraction.metafeature_extractor import BYU as MetafeatureExtractor

if __name__ == '__main__':
    dataframe = load_dataset()
    # dataframe['Class'] = np.NaN
    target_col_metadata = dict(dataframe.metadata.query((metadata_base.ALL_ELEMENTS,30)))
    # target_col_semantic_types = target_col_metadata.get('semantic_types', []) + ('https://metadata.datadrivendiscovery.org/types/RedactedTarget',)
    target_col_semantic_types = target_col_metadata.get('semantic_types', []) + ('https://metadata.datadrivendiscovery.org/types/TrueTarget',)
    # target_col_semantic_types = target_col_metadata.get('semantic_types', [])
    target_col_metadata['semantic_types'] = target_col_semantic_types
    dataframe.metadata = dataframe.metadata.update((metadata_base.ALL_ELEMENTS,30), target_col_metadata)
    hyperparams_class = MetafeatureExtractor.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
    hyperparams = hyperparams_class(metafeature_subset='INEXPENSIVE', metafeatures_to_compute=['linear_discriminant_analysis'])
    df_with_metafeatures = MetafeatureExtractor(hyperparams=hyperparams).produce(inputs=dataframe).value
    md = df_with_metafeatures.metadata.to_json_structure()
    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(md[0]['metadata']['data_metafeatures'])
    print('done testing d3m-loaded dataset')

    # FOR TESTING WITH JUST READ_CSV
    # dataset_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data/38_sick_dataset/tables/learningData.csv')
    # dataframe = DataFrame(pd.read_csv(dataset_path))
    # target_col_metadata = dict(dataframe.metadata.query((metadata_base.ALL_ELEMENTS,30)))
    # target_col_semantic_types = target_col_metadata.get('semantic_types', ()) + ('https://metadata.datadrivendiscovery.org/types/TrueTarget',)
    # target_col_metadata['semantic_types'] = target_col_semantic_types
    # dataframe.metadata = dataframe.metadata.update((metadata_base.ALL_ELEMENTS,30), target_col_metadata)
    # hyperparams_class = MetafeatureExtractor.metadata.query()['primitive_code']['class_type_arguments']['Hyperparams']
    # df_with_metafeatures = MetafeatureExtractor(hyperparams=hyperparams_class(metafeature_subset='ALL', metafeatures_to_compute=['class_probabilities', 'means_of_attributes'])).produce(inputs=dataframe).value
    # df_with_metafeatures.metadata.pretty_print(())
    # print('done testing pandas-loaded dataset')


import os

from d3m.container.dataset import Dataset
from d3m.container.pandas import DataFrame
print('importing common primitives package...')
from d3m.primitives.data_transformation.dataset_to_dataframe import Common as DatasetToDataFrame
from common_primitives.dataset_to_dataframe import Hyperparams as DatasetToDataFrameHyperparams
from d3m.primitives.data_transformation.column_parser import DataFrameCommon as ColumnParser
from common_primitives.column_parser import Hyperparams as ColumnParserHyperparams
print('done')

def load_dataset():
	dataset_doc_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data', '38_sick_dataset', 'datasetDoc.json'))
	dataset = Dataset.load(f'file://{dataset_doc_path}')
	dataframe = DatasetToDataFrame(hyperparams=DatasetToDataFrameHyperparams.defaults()).produce(inputs=dataset).value
	dataframe = ColumnParser(hyperparams=ColumnParserHyperparams.defaults()).produce(inputs=dataframe).value
	return dataframe


import json

from imputer import RandomSamplingImputer

outfile = 'jsons/imputer_primitive.json'
with open(outfile, 'w') as f:
	primitive_json = RandomSamplingImputer(hyperparams=None).metadata.to_json()
	f.write(json.dumps(primitive_json, indent=4, sort_keys=True))




from setuptools import setup, find_packages

__version__ = '0.6.0'

setup(
	name="byudml",
	packages = find_packages(),
	version = __version__,
	description = "A collection of DARPA D3M primitives developed by BYU",
	author = "Roland Laboulaye, Brandom Schoenfeld, Jarom Christensen",
	url = "https://github.com/byu-dml/d3m-primitives",
	include_package_data=True,
	keywords = ["metalearning", "metafeature", "machine learning", "metalearn", "d3m_primitive"],
	install_requires = [
		"metalearn==0.5.2",
		"numpy<=1.15.4",
        "scipy<1.2,>=0.19.0",
		"pandas<=0.23.4"
	],
	entry_points = {
    	'd3m.primitives': [
    		'data_preprocessing.random_sampling_imputer.BYU = byudml.imputer.random_sampling_imputer:RandomSamplingImputer',
    		'metafeature_extraction.metafeature_extractor.BYU = byudml.metafeature_extraction.metafeature_extraction:MetafeatureExtractor'
    	]
	}
)



from d3m.primitive_interfaces.base import CallResult
from d3m.primitive_interfaces.unsupervised_learning import UnsupervisedLearnerPrimitiveBase
from d3m import metadata
from d3m.metadata.hyperparams import Hyperparams
from d3m.metadata.params import Params
from d3m import container
import numpy as np
import pandas


__primitive_version__ = '0.1.3'
__package_version__ = '0.6.0'

Inputs = container.pandas.DataFrame
Outputs = container.pandas.DataFrame

class Params(Params):
    column_vals: container.list.List

class RandomSamplingImputer(UnsupervisedLearnerPrimitiveBase[Inputs, Outputs, Params, Hyperparams]):

    """
    A primitive which takes a DataFrame with "NaN" for all missing values, and imputes them for each column by randomly sampling from the existing values of that column.
    If a column has no existing values (aka a completely empty column), the column is ignored and remains in the dataset unimputed"
    """

    metadata = metadata.base.PrimitiveMetadata({
        'id': 'ebfeb6f0-e366-4082-b1a7-602fd50acc96',
        'version': __primitive_version__,
        'name': 'Random Sampling Imputer',
        'source': {
            'name': 'byu-dml',
            'contact': 'mailto:bjschoenfeld@gmail.com',
            'uris': [
                'https://github.com/byu-dml/d3m-primitives'
            ]
        },
        'installation': [
            {
                'type': metadata.base.PrimitiveInstallationType.PIP,
                'package': 'byudml',
                'version': __package_version__
            }
        ],
        'location_uris': [
            'https://github.com/byu-dml/d3m-primitives/blob/master/byu_dml/imputer/random_sampling_imputer.py'
        ],
        'python_path': 'd3m.primitives.data_preprocessing.random_sampling_imputer.BYU',
        'primitive_family': metadata.base.PrimitiveFamily.DATA_PREPROCESSING,
        'algorithm_types': [
            metadata.base.PrimitiveAlgorithmType.IMPUTATION
        ],
        'effects': [
            # not the case if empty columns are just ignored
            metadata.base.PrimitiveEffect.NO_MISSING_VALUES
        ]
    })

    def __init__(self, *, hyperparams: Hyperparams, random_seed: int=0) -> None:
        if random_seed == 0:
            random_seed = np.random.randint(100000)
        super().__init__(hyperparams=hyperparams, random_seed = random_seed)
        self._column_vals: container.list.List[container.list.List] = None
        self._random_state = np.random.RandomState(self.random_seed)
        self._training_inputs: Inputs = None
        self._fitted: bool = False

    def set_training_data(self, *, inputs: Inputs) -> None:
        self._training_inputs = inputs
        self._fitted = False

    def fit(self, *, timeout: float = None, iterations: int = None) -> CallResult[None]:
        if self._fitted:
            return CallResult(None)

        if self._training_inputs is None:
            raise ValueError('Missing training data.')

        self._column_vals = []
        dataframe = self._training_inputs
        for feature in dataframe:
            dropped_nan_series = dataframe[feature].dropna(axis=0, how='any')
            self._column_vals.append(dropped_nan_series)
        self._fitted = True

    def produce(self, *, inputs: Inputs, timeout: float = None, iterations: int = None) -> CallResult[Outputs]:
        if not self._fitted:
            raise ValueError("Calling produce before fitting.")
        
        dataframe = inputs
        for i in range(len(dataframe.columns)):
            feature_series = dataframe.iloc[:,i]
            col = feature_series.values
            num_nan = np.sum(feature_series.isnull())
            # ignores empty columns
            if len(self._column_vals[i]) > 0:
                col[feature_series.isnull()] = self._random_state.choice(self._column_vals[i], num_nan)
            else:
                self.logger.warning('\nWARNING: column labeld \'%s\' is completely empty - no values to impute.  This column contains no information and should be dropped by another primitive.', dataframe.columns[i])
        outputs = dataframe

        outputs.metadata = outputs.metadata.update((), {
            'schema': metadata.base.CONTAINER_SCHEMA_VERSION,
            'structural_type': type(outputs)
        })

        return CallResult(outputs)

    def get_params(self) -> Params:
        return Params(column_vals=self._column_vals)

    def set_params(self, *, params: Params) -> None:
        self._column_vals = params['column_vals']




